##############################################################################
# CEMFI SUMMER SCHOOL 2015
# PYTHON SCRIPT FOR E-ESTIMATION
# Bryan S. Graham, UC - Berkeley, bgraham@econ.berkeley.edu
# prepared by Bryan Graham & Seongjoo Min
# Last edited: 8/23/2015
##############################################################################

##############################################################################
# These codes are written for Python 2.7, using SciPy Stack package
# An installation for SciPy Stack package can be found at 
# http://www.scipy.org/install.html
# The .py script file and csv dat file need to be in the same folder
##############################################################################

#  Clears all variables
#  This code is borrowed from Christian A. Rossman, from a posting at
#  http://scientificcomputingco.blogspot.com/2013/10/
#  python-spyder-clear-workspaceshell.html

def clear_all():
    
    gl=globals().copy()
    
    for var in gl:
        if var[0]=='_': continue
        if 'func' in str(globals()[var]): continue
        if 'module' in str(globals()[var]): continue
        
        del globals()[var]
        
clear_all()

# Importing necessary modules

import pandas as pd
import numpy as np 
import statsmodels.api as sm
import statsmodels.sandbox.regression.gmm as gmm
import matplotlib.pyplot as plt
import scipy.stats as scs
import sys

##############################################################################
# Importing NLSY79 Extract (White Males)
##############################################################################

nlsy79 = pd.read_csv('nlsy79_wm.csv') # Reading .csv as DataFrame
nlsy79.set_index(['PID_79'], drop=False)

##############################################################################
# PARTIALLY LINEAR REGRESSION
##############################################################################

# Generate squares and cross products of elements of W
nlsy79['AFQT_sq']       = np.square(nlsy79.AFQT) 
nlsy79['HGCf_sq']       = np.square(nlsy79.HGC_FATH79r) 
nlsy79['HGCm_sq']       = np.square(nlsy79.HGC_MOTH79r) 
nlsy79['AFQT_x_HGCf']   = np.multiply(nlsy79.HGC_FATH79r,nlsy79.AFQT)  
nlsy79['AFQT_x_HGCm']   = np.multiply(nlsy79.HGC_MOTH79r,nlsy79.AFQT)  
nlsy79['HGCf_x_HGCm']   = np.multiply(nlsy79.HGC_FATH79r,nlsy79.HGC_MOTH79r)

y    = nlsy79['LogEarn'] 
X    = nlsy79['HGC_Age28']
X    = sm.add_constant(X) #Add constant to the design matrix
c_id = nlsy79['PID_79']

# Short regression: OLS fit of log earnings onto years of schooling
short_reg=sm.OLS(y,X).fit(cov_type='cluster', cov_kwds={'groups':c_id})
print "Short regression: no covariate adjustment"
print(short_reg.summary())

XW = nlsy79[['HGC_Age28','AFQT','HGC_FATH79r','HGC_MOTH79r']]
XW = sm.add_constant(XW)
# Long regression: OLS fit of log earnings onto years of schooling & controls
long_reg=sm.OLS(y,XW).fit(cov_type='cluster', cov_kwds={'groups':c_id})
print "Long regression: linear covariate adjustment"
print(long_reg.summary())

XkW = nlsy79[['HGC_Age28','AFQT','HGC_FATH79r','HGC_MOTH79r', \
              'AFQT_sq','HGCf_sq','HGCm_sq','AFQT_x_HGCf', \
              'AFQT_x_HGCm', 'HGCf_x_HGCm']]
XkW = sm.add_constant(XkW)
# PLM regression: OLS fit of log earnings onto years of schooling & polynomial
# in control variables
plm_reg=sm.OLS(y,XkW).fit(cov_type='cluster', cov_kwds={'groups':c_id})
print "Partially linear regression: semiparametric covariate adjustment"
print(plm_reg.summary())

##############################################################################
# COMPUTE GENERALIZED PROPENSITY SCORE
##############################################################################
X   = nlsy79['HGC_Age28']
kW  = nlsy79[['AFQT','HGC_FATH79r','HGC_MOTH79r', \
              'AFQT_sq','HGCf_sq','HGCm_sq','AFQT_x_HGCf', \
              'AFQT_x_HGCm', 'HGCf_x_HGCm']]
kW = sm.add_constant(kW)
gps = sm.GLM(X, kW, family=sm.families.Poisson())
gps_res = gps.fit()
print(gps_res.summary())

e_hat = gps_res.mu #fitted value  of e(W)

# Scatter plot to assess model fit
from statsmodels.graphics.api import abline_plot
fig, ax = plt.subplots()
ax.scatter(e_hat, X)
line_fit = sm.OLS(X, sm.add_constant(e_hat, prepend=True)).fit()
abline_plot(model_results=line_fit, ax=ax)

ax.set_title('Generalized Propensity Score')
ax.set_ylabel('X')
ax.set_xlabel('e(W)');

##############################################################################
# COMPUTE E-ESTIMATES
##############################################################################
Z = X - e_hat
beta_hat_E = np.dot(Z.T,y) / np.dot(Z.T,X)

# Calculate standard error
(N,J)=np.shape(kW) # dimension of k(W) basis function vector
s_phi = np.multiply(np.tile(Z,(J,1)).T,kW)   # N x J matrix of score contributions
m_hat = Z * (y - X*beta_hat_E)               # N x K (=1) matrix of m(Z,phi,beta) vectors
m_res = sm.OLS(m_hat,s_phi).fit().resid      # m - m_hat
Lambda = np.dot(Z.T,X)/N                       # Estimate of E[v(W)]
Omega  = np.dot(m_res,m_res)/N
beta_std_err = np.sqrt((Omega/(Lambda**2))/N)
print "---------------------------------"
print "E-Estimate of Return to Schooling"
print "---------------------------------"
print "E-Estimate of beta"
print beta_hat_E
print "Asymptotic standard error"
print beta_std_err
